// Copyright 2022, FOSS-VG Developers and Contributers
//
// Author(s):
//  BodgeMaster
//
// This program is free software: you can redistribute it and/or modify it
// under the terms of the GNU Affero General Public License as published
// by the Free Software Foundation, version 3.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
// See the GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// version 3 along with this program.
// If not, see https://www.gnu.org/licenses/agpl-3.0.en.html

#include <bitset>
#include <iomanip>
#include <iostream>
#include <vector>

#include "../lib/cli.hpp"
#include "../lib/file.hpp"

#define EXIT_SUCCESS 0
#define EXIT_RUNTIME 1
#define EXIT_USAGE   2

int main(int argc, char* argv[]) {
    std::vector<CLI::Flag> flags;
    flags.push_back(CLI::Flag('x', "hexadecimal", "write output as hexadecimal numbers"));
    flags.push_back(CLI::Flag('o', "octal",       "write output as octal numbers"));
    flags.push_back(CLI::Flag('b', "binary",      "write output as binary numbers"));
    flags.push_back(CLI::Flag('h', "help",        "print help and exit"));
    flags.push_back(CLI::Flag('l', "license",     "print license information and exit"));
    std::vector<CLI::Option> options;
    std::vector<CLI::Argument> arguments;
    arguments.push_back(CLI::Argument("FILE", "path of the file to dump"));
    CLI::ArgumentsParser cliParser = CLI::ArgumentsParser(argc, argv, flags, options, arguments, "Dump files as C++ array literals");

    if (cliParser.getFlag("help").value) {
        std::cout << cliParser.getUsage() << std::endl;
        return EXIT_SUCCESS;
    }

    if (cliParser.getFlag("license").value){
        std::cout
            << "Copyright 2022, FOSS-VG Developers and Contributers\n"
            << "\n"
            << "ArrayDump is part of the FOSS-VG development tool suite.\n"
            << "\n"
            << "This program is free software: you can redistribute it and/or modify it\n"
            << "under the terms of the GNU Affero General Public License as published\n"
            << "by the Free Software Foundation, version 3.\n"
            << "\n"
            << "This program is distributed in the hope that it will be useful,\n"
            << "but WITHOUT ANY WARRANTY; without even the implied\n"
            << "warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"
            << "See the GNU Affero General Public License for more details.\n"
            << "\n"
            << "You should have received a copy of the GNU Affero General Public License\n"
            << "version 3 along with this program.\n"
            << "If not, see https://www.gnu.org/licenses/agpl-3.0.en.html"
            << std::endl;
        return EXIT_SUCCESS;
    }

    if (cliParser.wrongUsage) {
        std::cout << cliParser.getUsage() << std::endl;
        return EXIT_USAGE;
    }

    ErrorOr<File*> filePointer = File::open(cliParser.getArgument(0).value, 'r');
    if (filePointer.isError) {
        std::cout << "Failed to open file: " << cliParser.getArgument(0).value << std::endl;
        return EXIT_RUNTIME;
    }
    File* file = filePointer.value;

    std::cout << "{";
    if (cliParser.getFlag("hexadecimal").value) {
        if (!file->eof()) {
            std::cout << "0x" << std::setw(2) << std::setfill('0') << std::hex << (int) file->readByte().value;
        }
        while (!file->eof()) {
            std::cout << ", 0x" << std::setw(2) << std::setfill('0') << std::hex << (int) file->readByte().value;
        }
    } else if (cliParser.getFlag("octal").value) {
        if (!file->eof()) {
            std::cout << "0" << std::setw(3) << std::setfill('0') << std::oct << (int) file->readByte().value;
        }
        while (!file->eof()) {
            std::cout << ", 0" << std::setw(3) << std::setfill('0') << std::oct << (int) file->readByte().value;
        }
    } else if (cliParser.getFlag("binary").value) {
        if (!file->eof()) {
            std::cout << "0b" << std::bitset<8>((int) file->readByte().value);
        }
        while (!file->eof()) {
            std::cout << ", 0b" << std::bitset<8>((int) file->readByte().value);
        }
    } else {
        if (!file->eof()) {
            std::cout << (int) file->readByte().value;
        }
        while (!file->eof()) {
            std::cout << ", " << (int) file->readByte().value;
        }
    }
    std::cout << "}" << std::endl;

    delete file;
    return EXIT_SUCCESS;
}
